
/* *************
   *** JAVAR ***
   ****************************************************
   *** Copyright (c) 1996/1997                      ***
   *** Fabian Breg               Indiana University *** 
   *** All Rights Reserved                          ***
   ****************************************************
   *** Please refer to the LICENSE file distributed ***
   *** with this software for further details on    ***
   *** the licensing terms and conditions.          ***
   ***                                              ***
   *** Please, report all bugs, comments, etc.      ***
   *** to: breg@extreme.indiana.edu                 ***
   ****************************************************
   *** nexusize.c : add object serialization methods
   ***
   ***
   *** Your courtesy in mentioning the use of this compiler 
   *** in any scientific work that presents results obtained
   *** by using (extensions or modifications of) this compiler
   *** is highly appreciated.
   ***
   *** NOTE! 
   *** This implementation does NOT provide a complete Java 
   *** front-end (unicode escapes are not supported
   *** and the semantic analysis is rather limited).
   ***
   *** */

/* ********************************************************
   *** INCLUDE FILES and DEFINITIONS                    ***
   ********************************************************  */

#include <assert.h>
#include "class.h"
#include "global.h"
#include "y.tab.h"

#define NEXUS_WRITE 0
#define NEXUS_READ 1
#define NEXUS_SIZEOF 2

/* Some symbol names that we are going to need.
   These will be initialized in nexusize_init()
*/
static symb_ptr *put_meths;
static symb_ptr *get_meths;
static symb_ptr *sizeof_meths;

static symb_ptr classs;
static symb_ptr Object;
static symb_ptr Serialize;
static symb_ptr SerObject;
static symb_ptr serializeWrite;
static symb_ptr serializeRead;
static symb_ptr serializeSizeof;
static symb_ptr size;
static symb_ptr SerialObject;
static symb_ptr writeObject;
static symb_ptr readObject;
static symb_ptr sizeofObject;
static symb_ptr writeFinalObject;
static symb_ptr readFinalObject;
static symb_ptr sizeofFinalObject;
static symb_ptr tmp;

/* Global variables */
static stmt_ptr cur_stmt;
static stmt_ptr nexusWriteMeth;
static stmt_ptr nexusReadMeth;
static stmt_ptr nexusSizeofMeth;
static expr_ptr package;

static int nexusable;
static int add_cons;
static int action;
static int loop_nest;

char *this_class;
char *super_class;


/* Check whether class is serializable, ie implements the Serializable interface.
   wrapper function for class loading function
*/
static int is_serializable(char *class)
{
    int res = implements(class, "java/io/Serializable");
    if(res)
	if(!strncmp("java/", class, 5))
	    res = 0;
    return res;
}

/* Check whether the given class is remote, ie implements the Remote interface.
   wrapper function for class loading function
*/
static int is_remote(char *class)
{
    return implements(class, "nexusrmi/Remote");
}

/* Convert bytecode representation for type to a number. */
static int get_type_id(char *type)
{
    switch(type[0])
    {
    case 'V' : return 0;
    case 'Z' : return 1;
    case 'B' : return 2;
    case 'S' : return 3;
    case 'I' : return 4;
    case 'J' : return 5;
    case 'C' : return 6;
    case 'F' : return 7;
    case 'D' : return 8;
    case 'L' : return 9;
    case '[' : return 10;
    }
    javar_out(-1, "Unexpected fall-through in get_type_id().");
    return -1;
}

/* Bytecode uses slashes (/) instead of dots (.) as a separater,
   so we replace these.
*/
static void replace_slash(char *s)
{
    int i;
    for(i = 0; i < strlen(s); i++)
	if(s[i] == '/')
	    s[i] = '.';
}

/* Get a string representation of the fully qualified name. qual is
   typically given a pointer to the package expression, while the
   name is obtained from the class header.
*/
static char *get_full_name(expr_ptr qual, char *name)
{
    char *s;
    char *q;
    int len;
    
    if(qual)
    {
	q = get_full_name(qual->u.name.lcomp, qual->u.name.id->symbol);
	len = strlen(q) + strlen(name);
	s = (char *)malloc((len + 2) * sizeof(char));
	sprintf(s, "%s.%s%c", q, name, '\0');
	free(q);
    }
    else
    {
	len = strlen(name);
	s = (char *)malloc((len + 1) * sizeof(char));
	strcpy(s, name);
	s[len] = '\0';
    }
    return s;
}

/* Return a pointer to the ast representation of the fully qualified name 'name'. */
static expr_ptr construct_name(char *name)
{
    expr_ptr res;
    char *q, *n, *t;
    int l;

    t = strrchr(name, '.');
    if(t)
    {
	t++;
	l = strlen(t);
	n = (char *)malloc((l + 1) * sizeof(char));
	strncpy(n, t, l);
	n[l] = '\0';
	l = strlen(name) - strlen(t);
	q = (char *)malloc((l) * sizeof(char));
	strncpy(q, name, l);
	q[l - 1] = '\0';
	res = make_name(construct_name(q), insert_table(n), 0);
	free(n);
	free(q);
    }
    else
	res = make_name(NULL, insert_table(name), 0);

    return res;
}

/* Return a pointer to the ast representation for the type given in
   bytecode notation.
*/
static type_ptr make_type(char *tname)
{
    type_ptr tp = NULL;
    char *t = strdup(tname);

    if(tname[0] == '[')
	tp = make_array_type(make_type(t + 1));
    else if(tname[0] == 'L')
    {
	char *cname = strdup(t + 1);
	cname[strlen(cname) - 1] = '\0';
	replace_slash(cname);
	tp = make_c_i_type(construct_name(cname));
	free(cname);
    }
    else
    {
	switch(t[0])
	{
	case 'Z':
	    tp = make_primitive_type(BOOL_T);
	    break;
	case 'B':
	    tp = make_primitive_type(BYTE_T);
	    break;
	case 'S':
	    tp = make_primitive_type(SHORT_T);
	    break;
	case 'I':
	    tp = make_primitive_type(INT_T);
	    break;
	case 'J':
	    tp = make_primitive_type(LONG_T);
	    break;
	case 'C':
	    tp = make_primitive_type(CHAR_T);
	    break;
	case 'F':
	    tp = make_primitive_type(FLOAT_T);
	    break;
	case 'D':
	    tp = make_primitive_type(DOUBLE_T);
	    break;
	}
    }
    free(t);
    return tp;
}

/* Generate a pointer to the ast representation for an array allocation expression.
   Returns an expression that looks like this: new [nr][][], where the number of
   dimensions is equal to the number of dimensions of tp.
*/
static expr_ptr array_alloc_expr(char *tp, symb_ptr nr)
{
    char *bt = tp;
    expr_ptr dim = make_op(']', make_op('[', NULL, make_name(NULL, nr, 0)), NULL);
    expr_ptr t = dim;
    type_ptr base_type;
    
    while(*bt == '[')
    {
	bt++;
	t->u.op.rhs = make_op('[', NULL, NULL);
	t = t->u.op.rhs;
    }

    if(get_type_id(bt) < 9)
	base_type = make_primitive_type(get_type_id(bt));
    else
    {
	char *elt_type = (char *)malloc((strlen(bt) - 1) * sizeof(char));
	strncpy(elt_type, bt + 1, strlen(bt) - 2);
	elt_type[strlen(bt) - 2] = '\0';
	replace_slash(elt_type);
        base_type = make_c_i_type(make_name(NULL, insert_table(elt_type), 0));
    }
    return make_te(NEW_, base_type, dim);
}

/* Produce type specific serialization code for the field with the given type and name. */
static stmt_ptr handle_type_expr(char *type, expr_ptr name)
{
    int type_id = get_type_id(type);
    stmt_ptr curstmt = NULL;

    switch(type_id)
    {
    case 0:
	assert(0); /* Should never happen */
	break;
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
    case 8:
	if(action == NEXUS_WRITE)
	    /* SerObject.write<primitive>(<name>); */
	    curstmt = make_simple(EXP_S,
				  make_op('i',
					  make_op('.',
						  make_name(NULL, SerObject, 0),
						  make_name(NULL, put_meths[type_id - 1], 0)),
					  name));
	else if(action == NEXUS_READ)
	    /* <name> = SerObject.read<primitive>(); */
	    curstmt = make_simple(EXP_S,
				  make_op('=',
					  name,
					  make_op('i',
						  make_op('.',
							  make_name(NULL, SerObject, 0),
							  make_name(NULL, get_meths[type_id - 1], 0)),
						  NULL)));
	else
  	    /* size += SerObject.sizeof<primitive>(1); */
	    curstmt = make_simple(EXP_S,
				  make_op(PIS_,
					  make_name(NULL, size, 0),
					  make_op('i',
						  make_op('.',
							  make_name(NULL, SerObject, 0),
							  make_name(NULL, sizeof_meths[type_id - 1], 0)),
						  name)));
	break;
    case 9:
    case 10:
	{
	    short final = 0;
	    char *cn = NULL;

	    if(type_id == 9)
	    {
		cn = strdup(type + 1);
		cn[strlen(cn) - 1] = '\0';
		final = is_final(cn);
	    }
	    
	    if(!final)
	    {
		if(action == NEXUS_WRITE)
		    /* SerObject.writeObject(<name>); */
		    curstmt = make_simple(EXP_S,
					  make_op('i',
						  make_op('.',
							  make_name(NULL, SerObject, 0),
							  make_name(NULL, writeObject, 0)),
						  name));
						      
		else if(action == NEXUS_READ)
		    /* <name> = (<type>)Nexusize.nexusReadObject(); */
		    curstmt = make_simple(EXP_S,
					  make_op('=',
						  name,
						  make_cast(make_type(type),
							    NULL,
							    make_op('i',
								    make_op('.',
									    make_name(NULL, SerObject, 0),
									    make_name(NULL, readObject, 0)),
								    NULL))));
		else
		    /* size += SerObject.sizeofObject(<name>); */
		    curstmt = make_simple(EXP_S,
					  make_op(PIS_,
						  make_name(NULL, size, 0),
						  make_op('i',
							  make_op('.',
								  make_name(NULL, SerObject, 0),
								  make_name(NULL, sizeofObject, 0)),
							  name)));
	    }
	    else
	    {
		replace_slash(cn);
/* 		char *cns = (char *)malloc((strlen(cn) + 3) * sizeof(char)); */
/* 		sprintf(cns, "\"%s\"", cn); */
		if(action == NEXUS_WRITE)
		    /* SerObject.writeFinalObject(<cn>, <name>); */
		    curstmt = make_simple(EXP_S,
					  make_op('i',
						  make_op('.',
							  make_name(NULL, SerObject, 0),
							  make_name(NULL, writeFinalObject, 0)),
						  make_op(',',
							  make_op('.',
								  construct_name(cn),
								  make_name(NULL, classs, 0)),
/* 							  make_literal(-1, cns, 0), */
							  name)));
						      
		else if(action == NEXUS_READ)
		    /* <name> = (<type>)SerObject.readFinalObject(<cn>); */
		    curstmt = make_simple(EXP_S,
					  make_op('=',
						  name,
						  make_cast(make_type(type),
							    NULL,
							    make_op('i',
								    make_op('.',
									    make_name(NULL, SerObject, 0),
									    make_name(NULL, readFinalObject, 0)),
								    make_op('.',
									    construct_name(cn),
									    make_name(NULL, classs, 0))))));
/* 								    make_literal(-1, cns, 0))))); */
		else
		    /* size += SerObject.sizeofFinalObject(<cn>, <name>); */
		    curstmt = make_simple(EXP_S,
					  make_op(PIS_,
						  make_name(NULL, size, 0),
						  make_op('i',
							  make_op('.',
								  make_name(NULL, SerObject, 0),
								  make_name(NULL, sizeofFinalObject, 0)),
							  make_op(',',
								  make_op('.',
									  construct_name(cn),
									  make_name(NULL, classs, 0)),
/* 								  make_literal(-1, cns, 0), */
								  name))));
	    }
	    free(cn);
	}
	break;
    }
    return curstmt;
}

/* This is a function that can be given to field_traverse(). It will call handle_type(). */
static void handle_field(fm_ptr f)
{
    stmt_ptr s;
    if(!(f->access_flags & (ACC_FINAL | ACC_STATIC | ACC_TRANSIENT)))
    {
	loop_nest = 0;
	s = handle_type_expr(get_typedescriptor(this_class, f), make_name(NULL, insert_table(get_name(this_class, f)), 0));
	cur_stmt = concat(cur_stmt, s);
    }
}

/* Call 'action' with every field in class 'class' */
static void field_traverse(char *class, void (*action)(fm_ptr))
{
    fm_ptr f;
    
    trav_class_init(class, 1, 1, 1);
    while(f = get_next_field())
	action(f);
}
			   
/* Add the nexusWrite() method to class s */
static void add_nexus_write(stmt_ptr s)
{
    nexusable = 0;

    if(!is_remote(this_class) && (is_serializable(this_class)))
    {
	cur_stmt = NULL;

	if(super_class && is_serializable(super_class))
	{
	    /* super.serializeWrite(SerObject); */
	    cur_stmt = concat(cur_stmt,
			      make_simple(EXP_S,
					  make_op('i',
						  make_op('.',
							  make_thissup(SUPER_),
							  make_name(NULL, serializeWrite, 0)),
						  make_name(NULL, SerObject, 0))));
	}
	else if(strcmp(super_class, "java/lang/Object"))
	{
	    javar_out(-1, "superobject of %s is not nexusable", this_class);
	}

	/* Handle every field */
	action = NEXUS_WRITE;
	field_traverse(this_class, handle_field);

	/* public void serializeWrite(nexusrmi.Serialize SerObject)
	   {
	       ...
	   }
	*/
	s->u.header.body = concat(s->u.header.body,
				  make_header(METHOD_S, PUB_B, serializeWrite, make_primitive_type(VOID_T), 0,
					      make_declexpr(give_m_t_c(0,
								       make_c_i_type(make_name(NULL, Serialize, TYPE_N)),
								       make_decl(SerObject, 0, NULL),
								       LOCPAR_N)),
					      NULL,
					      cur_stmt, 0));
    }
}

/* Add nexusRead() method to class s. */
static void add_nexus_read(stmt_ptr s)
{
    if((!is_remote(this_class)) && (is_serializable(this_class)))
    {
	cur_stmt = NULL;
	
	if(super_class && is_serializable(super_class))
	{
	    /* super.serializeRead(SerObject); */
	    cur_stmt = concat(cur_stmt,
			      make_simple(EXP_S,
					  make_op('i',
						  make_op('.',
							  make_thissup(SUPER_),
							  make_name(NULL, serializeRead, 0)),
						  make_name(NULL, SerObject, 0))));
	}

	/* handle every field */
	action = NEXUS_READ;
	field_traverse(this_class, handle_field);
	cur_stmt = concat(cur_stmt, make_simple(RET_S, make_thissup(THIS_)));

	/* public Object serializeRead(nexusrmi.Serialize SerObject)
	   {
	       ...
	   }
	*/
	s->u.header.body = concat(s->u.header.body,
				  make_header(METHOD_S, PUB_B, serializeRead, make_c_i_type(make_name(NULL, Object, 0)), 0,
					      make_declexpr(give_m_t_c(0,
								       make_c_i_type(make_name(NULL, Serialize, TYPE_N)),
								       make_decl(SerObject, 0, NULL),
								       LOCPAR_N)),
					      NULL,
					      cur_stmt, 0));
    }
}

/* Add nexusSizeof() method to class s */
static void add_nexus_sizeof(stmt_ptr s)
{
    if((!is_remote(this_class)) && (is_serializable(this_class)))
    {
	cur_stmt = NULL;
	
	if(is_serializable(super_class))
	{
	    /* size = super.serializeSizeof(SerObject); */
	    cur_stmt = concat(cur_stmt,
			      make_simple(EXP_S,
					  make_op('=',
						  make_name(NULL, size, 0),
						  make_op('i',
							  make_op('.',
								  make_thissup(SUPER_),
								  make_name(NULL, serializeSizeof, 0)),
							  make_name(NULL, SerObject, 0)))));
	}

	/* Handle every field */
	action = NEXUS_SIZEOF;
	field_traverse(this_class, handle_field);

	/* int size = 0; */
	cur_stmt = concat(give_m_t_c(0, make_primitive_type(INT_T), make_decl(size, 0, make_literal(INT_T, "0", 0)), LOCPAR_N), cur_stmt);
	/* return size; */
	cur_stmt = concat(cur_stmt, make_simple(RET_S, make_name(NULL, size, 0)));

	/* public int serializeSizeof(nexusrmi.Serialize SerObject)
	   {
	       ...
	   }
	*/
	s->u.header.body = concat(s->u.header.body,
				  make_header(METHOD_S, PUB_B, serializeSizeof, make_primitive_type(INT_T), 0,
					      make_declexpr(give_m_t_c(0,
								       make_c_i_type(make_name(NULL, Serialize, TYPE_N)),
								       make_decl(SerObject, 0, NULL),
								       LOCPAR_N)),
					      NULL, cur_stmt, 0));
    }
}

/* Add an empty constructor to class s */
static void add_empty_cons(stmt_ptr s)
{
    s->u.header.body = concat(s->u.header.body, make_header(CONS_S, PUB_B, s->u.header.name->u.name.id, NULL, 0, NULL, NULL, NULL, 0));
}

/* Check whether method s is an empty constructor. If it is make it public.
   This function is called for every method in a class to find out whether an empty constructor already exists.
*/
static void check_cons(stmt_ptr s)
{
    if((s->kind == CONS_S) && (s->u.header.e1 == NULL))
    {
	s->u.header.mod |= PUB_B; /* Make sure the empty constructor is public */
	add_cons = 0;
    }
}



/* driver routine */
static void add_nexus_meths(stmt_ptr s)
{
    if(s->kind == PCK_S)
	package = s->u.arg;
    
    if(s->kind == CLASS_S)
    {
	this_class = get_full_name(package, s->u.header.name->u.name.id->symbol);
	/* Read the class file corresponding to class s. */
	if(!read_classes(this_class))
	{
	    super_class = get_superclass(this_class);

	    /* Add the SerialObject interface */
	    if((!is_remote(this_class)) && (is_serializable(this_class)))
	    {
		s->u.header.e1 = make_te('t', make_c_i_type(make_name(NULL, SerialObject, 0)), s->u.header.e1);
		
		/* Add empty constructor if not already there. */
		add_cons = 1;
		stmt_traverse(s->u.header.body, check_cons, NULL);
		if(add_cons)
		    add_empty_cons(s);

		/* Add the serialization methods. */
		add_nexus_write(s);
		add_nexus_read(s);
		/* add_nexus_sizeof(s); */

		delete_classes();
	    }
	}
	free(this_class);
    }
}



/* Initialize
   ********** */
   
static void nexusize_init(void)
{
    /* Initialize the symbols needed. */
    classs = insert_table("class");
    Object = insert_table("Object");
    
    put_meths = (symb_ptr *)malloc(8 * sizeof(symb_ptr));
    put_meths[0] = insert_table("writeboolean");
    put_meths[1] = insert_table("writebyte");
    put_meths[2] = insert_table("writeshort");
    put_meths[3] = insert_table("writeint");
    put_meths[4] = insert_table("writelong");
    put_meths[5] = insert_table("writechar");
    put_meths[6] = insert_table("writefloat");
    put_meths[7] = insert_table("writedouble");

    get_meths = (symb_ptr *)malloc(8 * sizeof(symb_ptr));
    get_meths[0] = insert_table("readboolean");
    get_meths[1] = insert_table("readbyte");
    get_meths[2] = insert_table("readshort");
    get_meths[3] = insert_table("readint");
    get_meths[4] = insert_table("readlong");
    get_meths[5] = insert_table("readchar");
    get_meths[6] = insert_table("readfloat");
    get_meths[7] = insert_table("readdouble");

    sizeof_meths = (symb_ptr *)malloc(8 * sizeof(symb_ptr));
    sizeof_meths[0] = insert_table("sizeofboolean");
    sizeof_meths[1] = insert_table("sizeofbyte");
    sizeof_meths[2] = insert_table("sizeofshort");
    sizeof_meths[3] = insert_table("sizeofint");
    sizeof_meths[4] = insert_table("sizeoflong");
    sizeof_meths[5] = insert_table("sizeofchar");
    sizeof_meths[6] = insert_table("sizeoffloat");
    sizeof_meths[7] = insert_table("sizeofdouble");

    Serialize = insert_table("nexusrmi.Serialize");
    SerObject = insert_table("SerObject");
    serializeWrite = insert_table("serializeWrite");
    serializeRead = insert_table("serializeRead");
    serializeSizeof = insert_table("serializeSizeof");
    size = insert_table("size");
    SerialObject = insert_table("nexusrmi.SerialObject");
    writeObject = insert_table("writeObject");
    readObject = insert_table("readObject");
    sizeofObject = insert_table("sizeofObject");
    writeFinalObject = insert_table("writeFinalObject");
    readFinalObject = insert_table("readFinalObject");
    sizeofFinalObject = insert_table("sizeofFinalObject");

    tmp = insert_table("tmp");
}

/* The main routine. */
void nexusize(void)
{
    cur_stmt = NULL;
    nexusWriteMeth = NULL;
    nexusReadMeth = NULL;
    nexusSizeofMeth = NULL;
    package = NULL;
    
    nexusize_init();
    stmt_traverse(comp_unit, add_nexus_meths, NULL);
}
